% MATLAB code: Section 8.2 and Application
% File: TR_modified.m
% Coded by: Juwon Seo
%
% (* ::Package:: * The function file "createfigure4.m" is needed)
% The code is for the circulation density estimator with Gaussian kernel (CDest), 
% 95% confidence intervals for the circulation density estimator 
% using the local bootstrap for Markov processes by Paparoditis and Politis (2002) (BCI)
% and the p-value of the Test for Time reversibility test proposed by Beare
% and Seo (2014) (P_value_TRtest).
%
% The figure draws the circulation density estimator and the confidence
% intervals.
%
% References:
% Beare, B.K. and Seo, J. (2014).
%   Time-reversible copula-based Markov models.
%   Econometric Theory, 30(5), 923-960.
%   DOI: 10.1017/s0266466614000115.
% Paparoditis, E. and Politis, D.N. (2002). 
%   The local bootstrap for Markov processes.
%   Journal of Statistical Planning and Inference, 108(1-2), 301-328.
%   DOI: 10.1016/s0378-3758(02)00315-4.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% !!!!!!!!!!!! Import the data as a row vector : !!!!!!!!!!!!!!!!!!!!!! %
%% X00 = !!(ENTER THE DATA FILE NAME HERE)!!;    
% For example,
% X00 = markup';  
% X00: (1xT) univariate time series data   
 
T = length(X00')-1;  % Sample size: T=T-1 
% replace "markup" with the name of your own data file.
mboot = 10;    %  Pick the number of bootstrap samples (400 or 800)
P     = (0.05:0.05:0.95)';   % points : the domain of the circulation density (0,1) 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
X(1,:) = X00(:,1:T);  
Y(1,:) = X00(:,1:T);
X(2,:) = X00(:,2:T+1); 
Y(2,:) = X00(:,2:T+1); 
X1     = X(1,:);
X2     = X(2,:);             
             
%% Estimation %%
ran1 = min(X00');  
ran2 = max(X00');  
xi1  = ran1-(1/5)*(ran2-ran1);
xi2  = ran2+(1/5)*(ran2-ran1);

xi = linspace(xi1,xi2,200);  % linearly spaced vector

blankH = zeros(length(xi),length(xi));                   
blankF = ksdensity(X00,xi,'function','cdf');            
blankf = ksdensity(X00,xi);                            

% [1] circulation density estimator 
cd_est1 = zeros(length(P),length(P));    
cd_est2 = zeros(length(P),length(P));  

h = 1.06*std(X00)*(T^(-1/5)) ;  
% the bandwidth selection by Silverman's thumb rule (can be modified) 

for j1=1:length(P);
    blank_inv      = zeros(200,3);                                 
    blank_inv(:,1) = xi;
    blank_inv(:,2) = blankF-P(j1,1); 
    for j=1:200;
        if blank_inv(j,2) < -0.0005; 
           blank_inv(j,3) = 100;
           else 
           blank_inv(j,3) = blank_inv(j,2);
        end;
    end;
    blank_invF = sortrows(blank_inv,3);
    quasi_invF = blank_invF(1,1);                        
    quasi_invG = quasi_invF;

    DF1  = zeros(T,1);
    DF10 = zeros(T,1);
    for ii = 1:T;
        DF1(ii,1)  = (normpdf((quasi_invG-X(1,ii))/h))*(normcdf((quasi_invG-X(2,ii))/h));
        DF10(ii,1) = (normpdf((quasi_invG-X(1,ii))/h));
    end;
    num1   = (1/(T*h))*sum(DF1);
    denom1 = (1/(T*h))*sum(DF10);

    DF2  = zeros(T,1);
    DF20 = zeros(T,1);
    for ii = 1:T;
        DF2(ii,1)  = (normpdf((quasi_invG-X(2,ii))/h))*(normcdf((quasi_invG-X(1,ii))/h));
        DF20(ii,1) = (normpdf((quasi_invG-X(2,ii))/h));
    end;
    num2   = (1/(T*h))*sum(DF2);
    denom2 = (1/(T*h))*sum(DF20);

    cd_est1 = num1/denom1 ;                        
    cd_est2 = num2/denom2 ;               

    T_stat(j1,1) = (diag (cd_est2))'-(diag(cd_est1))';   % CD estimator 
%   The circulation density defined as the difference between the first
%   partial derivatives of the copula function.
            
    [vv, vid]  = min(abs(xi-quasi_invG));
    fFinv      = blankF (vid);
    avar(j1,1) = (1/(2*pi^(1/2)))*(1/fFinv)*(cd_est1+cd_est2-2*cd_est1*cd_est2);
end;  % end loop j1=1:length(P) with P the points of the circulation domain

% [2] TR test statistic
for e1=1:length (xi);
    for e2=1:length (xi);
        for i=1:T;
           blankH_0(i,1) = (X1(i)<=xi(e1)).*(X2(i)<=xi(e2));
        end;
        blankH(e1,e2) = mean(blankH_0);                           
    end;
end;

DblankH = blankH-blankH';
J_stat  = ((T+1)^(1/2))*max(abs(DblankH(:)));              % TR test statistic

% [3] PP test statistic (lagx1 vector) lag=1,2,3,...,10]
for lag=1:10;
    Dlag        = X00(:,lag+1:T+1)-X00(:,1:T-(lag-1));
    I_Dlag      = Dlag<=0;
    Phat(lag,1) = mean(I_Dlag);                          
end;
V_stat = Phat;

%% Bootstrap Pseudo Samples SX            
% optimal bandwidth choice b: with Gaussian kernel and AR (1) process
% proposed in the paper by Paparoditis and Politis
     mat0 = X2';                                
     mu   = mean(X00);
     T_p  = std(X00)^2;
     ols  = [ones(1,T)' X(1,:)']\(X(2,:)');
     regd = X2'-ols(1)*ones (1,T)'-ols(2)*X1';    % residuals
     w1   = 1/(2*pi^(1/2));                       % Gaussian kernel
     c1   = ols(2,1)*(1/T_p)*(mat0-mu);
     c2   = ols(2,1)^2;
     sig  = (1/(T-1))*sum(regd.^2);               % residual variance

     for iii=1:T;                               
         [minv,id]     = min(abs(xi-X00(iii)));
         X00_ap(iii,1) = id;
     end; 
     b = ((sig^4*w1)./(T*blankf(X00_ap).*(2*sig^2*c1.^2+0.25*c2^2*ones(T,1))')).^(1/5);   % w2=1
     % see eq. (13) in PP (2002, p. 315); optimal resampling width; blankf = kernel density 
        
for m=1:mboot;      %  start bootstrap loop 
%     m   
     % Transition matrix
     for itm=1:T;
        for jtm=1:T;
           TM0(itm,jtm) = (1/b(itm))*normpdf((X00(itm+1)-X00(jtm))/b(itm)); % transition weight
        end;
     end;
     rowsum = sum(TM0')';
     for ltm=1:T;
        TM(ltm,:) = TM0(ltm,:)./rowsum(ltm);  % sum (TM')=1:row sum=1   transition matrix
     end;    
     CTM = cumsum(TM,2);                      % cumulative TM: Aprob
   
     % resampling
     SX00      = zeros(T,1); 
     o1        = randperm(T);
     o2        = o1(1,1);       % random permutation
     SX00(1,1) = X00(o2+1); 
     r0        = o2;            % initial value: random pick among X_2~~X_T
     for iib=2:T+1;
         r = rand();
         ind_SX00    = find(r < CTM (r0,:), 1, 'first');
         SX00(iib,1) = X00(ind_SX00+1); 
         r0          = ind_SX00;     
     end;
     SX00    = SX00';
     SX(1,:) = SX00(:,1:T);
     SX(2,:) = SX00(:,2:T+1);     
     SX1     = SX(1,:); 
     SX2     = SX(2,:);
          
     % [1] TR test statistic - pseudo
     for e1=1:length(xi);
         for e2=1:length(xi);
             for i=1:T;
                 SblankH_0(i,1) = (SX1(i)<=xi(e1)).*(SX2(i)<=xi(e2));
             end;
             SblankH(e1,e2)     = mean(SblankH_0);                         
         end;
     end;
     SDblankH       = SblankH-SblankH';
     SJ_stat10(:,m) = SDblankH(:);


     % [2] PP test statistic (lagx1 vector) lag=1,2,3,...,10 - pseudo
     for lag=1:10;
         SDlag        = SX00(:,lag+1:T+1)-SX00(:,1:T-(lag-1));
         SI_Dlag      = (SDlag<=0);
         SPhat(lag,1) = mean(SI_Dlag);                          
     end;
     SV_stat0 = SPhat;

     % [3] CD estimator- pseudo
     SblankF  = ksdensity(SX00,xi,'function','cdf');           
     Sblankf  = ksdensity(SX00,xi);                            
     Scd_est1 = zeros(length(P),length(P));    
     Scd_est2 = zeros(length(P),length(P));  
     Sh       = 1.06*std(SX00)*(T^(-1/5)) ;  

     for j1=1:length(P);
         Sblank_inv      = zeros(200,3);                                 
         Sblank_inv(:,1) = xi;
         Sblank_inv(:,2) = SblankF-P(j1,1); 
         for j=1:200;
           if Sblank_inv(j,2) < -0.0005; 
               Sblank_inv(j,3) = 100;
             else 
               Sblank_inv(j,3) = Sblank_inv(j,2);
           end;
         end;
         Sblank_invF = sortrows(Sblank_inv,3);
         Squasi_invF = Sblank_invF(1,1);                        
         Squasi_invG = Sblank_invF(1,1);

         SDF1  = zeros(T,1);
         SDF10 = zeros(T,1);
         for ii = 1:T;
            SDF1(ii,1)  = (normpdf((Squasi_invG-SX(1,ii))/Sh))*(normcdf((Squasi_invG-SX(2,ii))/Sh));
            SDF10(ii,1) = (normpdf((Squasi_invG-SX(1,ii))/Sh));
         end;
         Snum1   = (1/(T*h))*sum(SDF1);
         Sdenom1 = (1/(T*h))*sum(SDF10);

         SDF2    = zeros(T,1);
         SDF20   = zeros(T,1);
         for ii = 1:T;
            SDF2(ii,1)  = (normpdf((Squasi_invG-SX(2,ii))/Sh))*(normcdf((Squasi_invG-SX(1,ii))/Sh));
            SDF20(ii,1) = (normpdf((Squasi_invG-SX(2,ii))/Sh));
         end;
         Snum2    = (1/(T*h))*sum(SDF2);
         Sdenom2  = (1/(T*h))*sum(SDF20);
         Scd_est1 = Snum1/Sdenom1;                      
         Scd_est2 = Snum2/Sdenom2;                      

         ST_stat(j1,1) = (diag(Scd_est2))'-(diag(Scd_est1))';
     end;
     STstat0(m,:) = ST_stat'; 
end;   %  end bootstrap loop


for i=1:mboot;
   SJ_stat11(:,i)= SJ_stat10(:,i)-mean(SJ_stat10')';
end;
SJstat1 = (((T+1)^(1/2))*max(abs(SJ_stat11)))';

J1cvmat  = sortrows(SJstat1);           
J1cv95_1 = J1cvmat(mboot*(95/100),:); 
J1pvalue = mean(J1cvmat-abs (J_stat) > 0);

Tmean = mean(STstat0);
for i=1:length (P);    
    STstat(:,i) = (((T+1)*h)^(1/2))*(STstat0(:,i)-Tmean(i));           
end;

for i=1:length(P);      
    Tcv          = sortrows(STstat (:,i));
    Tcv95_1(i,1) = Tcv(mboot*(2.5/100),:);
    Tcv95_2(i,1) = Tcv(mboot*(97.5/100),:);  
end;     

STsig = std(STstat);                                                      
for i=1:length (P);      
    Tcv95_3(i,:) = T_stat(i)-1.96*STsig(1,i)*(1/(h*(T+1))^(1/2));                 
    Tcv95_4(i,:) = T_stat(i)+1.96*STsig(1,i)*(1/(h*(T+1))^(1/2));             
end;  

stat = T_stat;
BCI  = [T_stat+(((T+1)*h)^(-1/2))*Tcv95_1 T_stat+(((T+1)*h)^(-1/2))*Tcv95_2];

%% RESULTS
% Tcvs = [Tcv95_1 Tcv95_2]; Tcvs2=[Tcv95_3 Tcv95_4];  
% aci  = [T_stat-1.96*(avar.^(1/2))./((T+1)*h)^(1/2) T_stat+1.96*(avar.^(1/2))./((T+1)*h)^(1/2)];

P_value_TRtest = J1pvalue
CDest = T_stat
BCI

%% createfigure4(P', [stat BCI])   % optional






